@@ -15,7 +15,3 @@ MODULE_SCOPE int TkimgTIFFInitPixar(TIFF *, int);
 MODULE_SCOPE void TkimgTIFFfree(tdata_t data);
 MODULE_SCOPE tdata_t TkimgTIFFmalloc(tsize_t size);
 MODULE_SCOPE tdata_t TkimgTIFFrealloc(tdata_t data, tsize_t size);
-
-#ifndef CONST86
-#   define CONST86
-#endif
